/*
 * hostapd / Wi-Fi Simple Configuration 7C Proposal
 * Copyright (c) 2005 Intel Corporation. All rights reserved.
 * Contact Information: Harsha Hegde  <harsha.hegde@intel.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Alternatively, this software may be distributed under the terms of BSD
 * license.
 *
 * See README, README_WSC and COPYING for more details.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>

#include "eloop.h"
#include "hostapd.h"
#include "udplib.h"
#include "beacon.h"
#include "wsc.h"

#define WSC_INIT_READ_TIMEOUT	30 

/*

// Comment out the next line if debug strings are not needed
#define U_DEBUG

void DEBUGF(char *, ...);
*/


struct wsc_ie_data {
        struct hostapd_data *hapd;
        int udp_fd;
};

static struct wsc_ie_data *g_wsc_data;


#define DATA_ELEM_DUMP_NONE	1
#define DATA_ELEM_DUMP_OCTAL	2
#define DATA_ELEM_DUMP_STR	3
	
struct data_elem_def {
	u16 id;
	u16 len;
	u16 dumptype;
	char *desc;
};

static struct data_elem_def g_data_elem_def[] = {
	{	0x1001, 	2, 	DATA_ELEM_DUMP_OCTAL,	"AP Channel"	},
	{	0x1002,		2, 	DATA_ELEM_DUMP_OCTAL,	"Association State" },
	{	0x1003,		2, 	DATA_ELEM_DUMP_OCTAL,	"Authentication Type" },
	{	0x1004,		2,	DATA_ELEM_DUMP_OCTAL,	"Authentication Type Flags" },
	{	0x1005,		8,	DATA_ELEM_DUMP_OCTAL,	"Authenticator" },
	{	0x1008,		2,	DATA_ELEM_DUMP_OCTAL,	"Config Method" },
	{	0x1009,		2,	DATA_ELEM_DUMP_OCTAL,	"Configuration Error" },
	{	0x100A,		64,	DATA_ELEM_DUMP_STR,	"Confirmation url4" },
	{	0x100B,		76,	DATA_ELEM_DUMP_STR,	"Confirmation url6" },
	{	0x100C,		1,	DATA_ELEM_DUMP_OCTAL,	"Connection type" },
	{	0x100D, 	1,	DATA_ELEM_DUMP_OCTAL,	"Connection type flags" },
	{	0x100D,		0xFFFF,	DATA_ELEM_DUMP_OCTAL,	"Credential" },
	{	0x1011,		32, 	DATA_ELEM_DUMP_STR,	"Device name" },
	{	0x1012, 	2, 	DATA_ELEM_DUMP_OCTAL,	"Device Password ID" },
	{	0x1014,		32, 	DATA_ELEM_DUMP_OCTAL,	"E-Hash1" },
	{	0x1015,		32,	DATA_ELEM_DUMP_OCTAL,	"E-Hash2" },
	{	0x1016,		16,	DATA_ELEM_DUMP_OCTAL,	"E-SNonse1" },
	{	0x1017,		16,	DATA_ELEM_DUMP_OCTAL,	"E-SNonse2" },
	{	0x1018,		0xFFFF,	DATA_ELEM_DUMP_NONE,	"Encrypted settings" },
	{	0x100F,		2,	DATA_ELEM_DUMP_OCTAL,	"Encryption type" },
	{	0x1010,		2,	DATA_ELEM_DUMP_OCTAL,	"Encryption type flags" }, 
	{	0x101A,		16,	DATA_ELEM_DUMP_OCTAL,	"Enrollee nonse" },
	{	0x101B,		4,	DATA_ELEM_DUMP_OCTAL,	"Feature ID" },
	{	0x101C,		80,	DATA_ELEM_DUMP_STR,	"Identity" },
	{	0x101D,		0xFFFF,	DATA_ELEM_DUMP_NONE,	"Identity Proof" },
	{	0x101E,		8,	DATA_ELEM_DUMP_OCTAL,	"Key wrap authenticator" },
	{	0x101F,		16,	DATA_ELEM_DUMP_OCTAL,	"Key identifier" },
	{	0x1020,		6,	DATA_ELEM_DUMP_OCTAL,	"Mac Address" },
	{	0x1021,		64,	DATA_ELEM_DUMP_STR,	"Manufacturer" },
	{	0x1022,		1,	DATA_ELEM_DUMP_OCTAL,	"Message Type" },
	{	0x1023,		32,	DATA_ELEM_DUMP_STR,	"Model Name" },
	{	0x1024,		32,	DATA_ELEM_DUMP_OCTAL,	"Model number" },
	{	0x1026,		1,	DATA_ELEM_DUMP_OCTAL,	"Network Index" },
	{	0x1027,		64,	DATA_ELEM_DUMP_OCTAL,	"Network Key" },
	{	0x1028,		1,	DATA_ELEM_DUMP_OCTAL,	"Network Key Index" },
	{	0x1029,		32,	DATA_ELEM_DUMP_STR,	"New Device Name" },
	{	0x102A,		64,	DATA_ELEM_DUMP_NONE,	"New Password" },
	{	0x102C,		58,	DATA_ELEM_DUMP_NONE,	"OOB Device Password" },
	{	0x102D,		4,	DATA_ELEM_DUMP_OCTAL,	"OS Version" },
	{	0x102F,		1,	DATA_ELEM_DUMP_OCTAL,	"Power Level" },
	{	0x1030,		1,	DATA_ELEM_DUMP_OCTAL,	"PSK Current" },
	{	0x1031, 	1,	DATA_ELEM_DUMP_OCTAL,	"PSK Max" },
	{	0x1032,		192,	DATA_ELEM_DUMP_NONE, 	"Public Key" },
	{	0x1033,		1,	DATA_ELEM_DUMP_OCTAL,	"Radio Enabled" },
	{	0x1034,		1,	DATA_ELEM_DUMP_OCTAL,	"Reboot" },
	{	0x1035,		1,	DATA_ELEM_DUMP_OCTAL,	"Registrar Current" },
	{	0x1036,		1,	DATA_ELEM_DUMP_OCTAL,	"Registrar Established" },
	{	0x1037,		512,	DATA_ELEM_DUMP_OCTAL,	"Registrar List" },
	{	0x1038,		1,	DATA_ELEM_DUMP_OCTAL,	"Registrar Max" },
	{	0x1039,		16,	DATA_ELEM_DUMP_OCTAL,	"Registrar Nonce" },
	{	0x103A,		1,	DATA_ELEM_DUMP_OCTAL,	"Request Type" },
	{	0x103B,		1,	DATA_ELEM_DUMP_OCTAL,	"Response Type" },
	{	0x103C,		1,	DATA_ELEM_DUMP_OCTAL,	"RF Bands" },
	{	0x103D,		32,	DATA_ELEM_DUMP_OCTAL,	"R-Hash1" },
	{	0x103E,		32,	DATA_ELEM_DUMP_OCTAL,	"R-Hash2" },
	{	0x103F,		16,	DATA_ELEM_DUMP_OCTAL,	"R-SNonce1" },
	{	0x1040,		16,	DATA_ELEM_DUMP_OCTAL,	"R-SNonce2" },
	{	0x1041,		1,	DATA_ELEM_DUMP_OCTAL,	"Selected Registrar" },
	{	0x1042,		32,	DATA_ELEM_DUMP_OCTAL,	"Serial Number" },
	{	0x1044,		1,	DATA_ELEM_DUMP_OCTAL,	"Simple Config State" },
	{	0x1045,		32,	DATA_ELEM_DUMP_STR,	"SSID" },
	{	0x1046,		1,	DATA_ELEM_DUMP_OCTAL,	"Total Networks" },
	{	0x1047,		16,	DATA_ELEM_DUMP_OCTAL,	"UUID-E" },
	{	0x1048,		16,	DATA_ELEM_DUMP_OCTAL,	"UUID-R" },
	{	0x1049,		1024,	DATA_ELEM_DUMP_OCTAL,	"Vendor Extension" },
	{	0x104A,		1,	DATA_ELEM_DUMP_OCTAL,	"Version" },
	{	0x104B,		0xFFFF,	DATA_ELEM_DUMP_NONE,	"X.509 Certificate Request" },
	{	0x104C,		0xFFFF,	DATA_ELEM_DUMP_NONE,	"X.509 Certificate" },
	{	0x104D,		64,	DATA_ELEM_DUMP_OCTAL,	"EAP Identity" },
	{	0x104E,		8,	DATA_ELEM_DUMP_OCTAL,	"Message Counter" },
	{	0x104F,		20,	DATA_ELEM_DUMP_OCTAL,	"Public Key Hash" },
	{	0x1050,		32,	DATA_ELEM_DUMP_OCTAL,	"Rekey Key" },
	{	0x1051,		4,	DATA_ELEM_DUMP_OCTAL,	"Key Lifetime" },
	{	0x1052,		2,	DATA_ELEM_DUMP_OCTAL,	"Permitter Config Methods" },
	{	0x1053,		2,	DATA_ELEM_DUMP_OCTAL,	"Selected Registrar Config Methods" },
	{	0x1054,		8,	DATA_ELEM_DUMP_OCTAL,	"Primary Device Type" },
	{	0x1055,		128,	DATA_ELEM_DUMP_OCTAL,	"Secondary Device Type List" },
	{	0x1056,		1,	DATA_ELEM_DUMP_OCTAL,	"Portable Device" },
	{	0x1057,		1,	DATA_ELEM_DUMP_OCTAL,	"AP Setup Locked" },
	{	0x1058,		512,	DATA_ELEM_DUMP_OCTAL,	"Application Extension" },
	{	0x1059,		8,	DATA_ELEM_DUMP_OCTAL,	"EAP Type" },
	{	0x1060,		32,	DATA_ELEM_DUMP_OCTAL,	"Initialization Vector" },
	{	0x1061,		1,	DATA_ELEM_DUMP_OCTAL,	"Key Provided Automaticaly" },
	{	0x1062,		1,	DATA_ELEM_DUMP_OCTAL,	"802.1X Enabled" },
	{	0x1063,		128,	DATA_ELEM_DUMP_OCTAL,	"AppSessionKey" },
	{	0x1064,		1,	DATA_ELEM_DUMP_OCTAL,	"WEPTransmitKey" },

};

#define DATA_ELEM_DEF_SZ	(sizeof(g_data_elem_def)/sizeof(struct data_elem_def))  

static struct data_elem_def *get_data_elem_def_by_id(u16 id)
{
	int i;
	for (i = 0; i < DATA_ELEM_DEF_SZ; i++) {
		if (g_data_elem_def[i].id == id)
			return &(g_data_elem_def[i]);
	}
	return NULL;
}
 
void wsc_dump_ie(char *buf, int len)
{
	WSC_IE *ie = (WSC_IE*)buf;
	int pos, rem;
	u16 eid, elen;
	struct data_elem_def *elem;

	if (buf == NULL || len < 2) {
		wpa_printf(MSG_ERROR, "%s: IE buffer is NULL or is too small", __FUNCTION__);
		return;
	}

	wpa_printf(MSG_ERROR, "%s: elem id=0x%X length=%d", __FUNCTION__, ie->hdr.elemId, ie->hdr.length);
	
	if (ie->hdr.elemId != 221 || ie->hdr.length < 4) {
		wpa_printf(MSG_ERROR, "%s: Not WSC IE", __FUNCTION__);
		return;
	}

	wpa_printf(MSG_ERROR, "%s: OUI=0x%X 0x%X 0x%X 0x%X", __FUNCTION__, ie->hdr.oui[0], ie->hdr.oui[1], ie->hdr.oui[2], ie->hdr.oui[3]);
	
	for (pos = 6, rem = (ie->hdr.length - 4); rem > 0; ) {
		if (rem < 4) {
			wpa_printf(MSG_ERROR, "%s: remained less than 4 bytes in the buffer!", __FUNCTION__);
			return;
		}
		eid = buf[pos++] << 8;
		eid += buf[pos++];
		elen = buf[pos++] << 8;
		elen += buf[pos++];
		rem -= 4;

		elem = get_data_elem_def_by_id(eid);
		if (elem == NULL) {
			wpa_printf(MSG_ERROR, "%s: unknown data elem id 0x%X", __FUNCTION__, eid);
			return;
		}
		wpa_printf(MSG_ERROR, "%s: data elem id=0x%X (%s) length=%d", __FUNCTION__, eid, elem->desc, elen);
		wpa_hexdump_ascii(MSG_ERROR, elem->desc, &buf[pos], elen);
		pos += elen;
		rem -= elen;
	}  	
}

// set the Simple Config IE for beacon
u8 * hostapd_eid_wsc_beacon(struct hostapd_data *hapd, u8 *eid)
{
	u8 *pos = eid;
	if (IS_WSC_BEACON_IE_ENABLED(hapd->conf->wsc_flags)) {
		u8 size = sizeof(WSC_IE_HEADER)+ hapd->conf->wsc_ie_beacon.hdr.length-4;
		memcpy(pos, &hapd->conf->wsc_ie_beacon , size );
 		pos += size;
		//wpa_printf(MSG_ERROR,  "%s", __FUNCTION__);
		//wsc_dump_ie(&hapd->conf->wsc_ie_beacon, size);
	}
	return pos;
}


// set the Simple Config IE for Probe Response
u8 * hostapd_eid_wsc_probe_resp(struct hostapd_data *hapd, u8 *eid)
{
	u8 *pos = eid;
	if (IS_WSC_PROBE_RESP_IE_ENABLED(hapd->conf->wsc_flags)) {
		u8 size = sizeof(WSC_IE_HEADER) + hapd->conf->wsc_ie_probe_resp.hdr.length-4;
		memcpy(pos, &hapd->conf->wsc_ie_probe_resp , size);
 		pos += size;
		// wpa_printf(MSG_ERROR,  "%s", __FUNCTION__);
		// wsc_dump_ie(&hapd->conf->wsc_ie_probe_resp, size);
	}
	return pos;
}


static int  wsc_process_command(struct wsc_ie_data *data, char *buf, int len)
{
	WSC_IE_COMMAND_DATA *cmdData;
	u8 *bufPtr;

	if (len < sizeof (WSC_IE_COMMAND_DATA))
		return -1;

	cmdData = (WSC_IE_COMMAND_DATA*) buf;
	bufPtr = (u8*) cmdData->data;

	if ((len - sizeof(WSC_IE_COMMAND_DATA)) < cmdData->length) {
		wpa_printf(MSG_ERROR, "%s: Invalid ", __FUNCTION__);
		return -1;
	}

	switch (cmdData->type) {
		case WSC_IE_TYPE_SET_BEACON_IE:
			wpa_printf(MSG_INFO, "%s: SET_BEACON_IE from upper layer\n", __FUNCTION__);
			data->hapd->conf->wsc_flags |= WSC_BEACON_IE_ENABLED;	
			memcpy(&data->hapd->conf->wsc_ie_beacon, bufPtr, cmdData->length);
			ieee802_11_set_beacon(data->hapd);
			wpa_hexdump(MSG_ERROR, "WSC_IE_TYPE_SET_BEACON_IE", bufPtr, cmdData->length);
			break;
		case WSC_IE_TYPE_SET_PROBE_RESPONSE_IE:
			wpa_printf(MSG_INFO, "%s: SET_PR_RESP_IE from upper layer\n", __FUNCTION__);
			data->hapd->conf->wsc_flags |= WSC_PROBE_RESP_IE_ENABLED;
			memcpy(&data->hapd->conf->wsc_ie_probe_resp, bufPtr, cmdData->length);
			wpa_hexdump(MSG_ERROR, "WSC_IE_TYPE_SET_PROBE_RESPONSE_IE", bufPtr, cmdData->length);
			break;
		default:
			wpa_printf(MSG_ERROR, "%s: Wrong command type from upper layer\n", __FUNCTION__);
	}
	
	return 0;
}

static void wsc_ie_read_callback(int sock, void *eloop_ctx, void *sock_ctx)
{
	struct wsc_ie_data *data = eloop_ctx;
	char readBuf[WSC_WLAN_DATA_MAX_LENGTH];
	int recvBytes;
	struct sockaddr_in from;

	wpa_printf(MSG_INFO, "WSC_IE: Entered wsc_ie_read_callback. sock = %d", sock);

	recvBytes = udp_read(data->udp_fd, readBuf, WSC_WLAN_DATA_MAX_LENGTH, &from);

	if (recvBytes == -1) {
        	wpa_printf(MSG_ERROR, "WSC_IE: Reading Command message from upper layer failed");
        	return;
	}

	wsc_process_command(data, readBuf, recvBytes);
}

int wsc_init(struct hostapd_data *hapd)
{
	struct sockaddr_in addr;
	int bytes;
	char buf[] = "PORT";
#define PORT_MSG_SZ sizeof(buf)
	char inbuf[WSC_WLAN_DATA_MAX_LENGTH] = {0};

	wpa_printf(MSG_INFO, "WSC_IE: In wsc_ie_init\n");

	g_wsc_data = malloc(sizeof(struct wsc_ie_data));
	if (g_wsc_data == NULL) 
		return -1;

	memset(g_wsc_data, 0, sizeof(struct wsc_ie_data));

	g_wsc_data->hapd = hapd;
	g_wsc_data->udp_fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (g_wsc_data->udp_fd < 0) {
		wpa_printf(MSG_ERROR, "WSC_IE: Failed to open DGRAM socket\n");
		return -1;
	}
	
	addr.sin_addr.s_addr = inet_addr(WSC_WLAN_UDP_ADDR);
	addr.sin_family = AF_INET;
	addr.sin_port = htons(WSC_WLAN_UDP_PORT);

	bytes = udp_write(g_wsc_data->udp_fd, buf,  PORT_MSG_SZ, &addr); 
	if (bytes != PORT_MSG_SZ) {
		wpa_printf(MSG_ERROR, "WSC_IE: Sending Port message to upper Layer failed");
        	return -1;
	}

	// get initial beacon and probe response wsc ies
	while (	!IS_WSC_BEACON_IE_ENABLED(g_wsc_data->hapd->conf->wsc_flags) || 
		!IS_WSC_PROBE_RESP_IE_ENABLED(g_wsc_data->hapd->conf->wsc_flags)) {
		bytes = udp_read_timed(g_wsc_data->udp_fd, inbuf, WSC_WLAN_DATA_MAX_LENGTH, &addr, WSC_INIT_READ_TIMEOUT);	
		wsc_process_command(g_wsc_data, inbuf, bytes);
		if (bytes < 0) {
			wpa_printf(MSG_ERROR, "WSC_IE: Timeout on WSC sync reply message.\n");
	                return -1;
		}
	} 

	eloop_register_read_sock(g_wsc_data->udp_fd, wsc_ie_read_callback, g_wsc_data, NULL);

	hapd->conf->wsc_flags |= WSC_ENABLED;

	return 0;
}


int wsc_deinit(struct hostapd_data *hapd)
{
	struct wsc_ie_data *ptr = g_wsc_data;

	if (g_wsc_data == NULL)
		return 0;

	g_wsc_data = NULL;
	ptr->hapd->conf->wsc_flags = 0;
	if (ptr->udp_fd > -1) {
		eloop_unregister_read_sock(ptr->udp_fd);
		udp_close(ptr->udp_fd);
	}

	free(ptr);

	return 0;
}



